@@ -7,8 +7,9 @@ import inspect
 import logging
 import traceback
 import shlex
-from dpyConsole.errors import CommandNotFound
+import threading
 
+from dpyConsole.errors import CommandNotFound, ExtensionError
 
 logger = logging.getLogger("dpyConsole")
 
@@ -25,10 +26,19 @@ class Console:
         self.out = kwargs.get("out", sys.stdout)
         self.__commands__ = dict()
         self.converter = kwargs.get("converter", Converter(client))
+        self.__extensions = {}
+        self.__cogs = {}
 
     def add_console_cog(self, obj):
         if isinstance(obj, Cog):
-            self.__commands__.update(obj.commands)
+            self.__cogs.update({obj.__class__.__name__: obj})
+            obj.load(self)
+            return
+        raise Exception
+
+    def remove_console_cog(self, name):
+        cog = self.__cogs.pop(name, None)
+        cog.unload(self)
 
     def load_extension(self, path):
         """
@@ -36,9 +46,52 @@ class Console:
         :param path:
         :return:
         """
+        if path in self.__extensions:
+            raise ExtensionError(f"Extension {path} already loaded")
         module = importlib.import_module(path)
         # noinspection PyUnresolvedReferences
         module.setup(self)
+        self.__extensions.update({path: module})
+
+    def unload_extension(self, path):
+        """
+        Unloads an extension
+        :param path:
+        :return:
+        """
+        module = self.__extensions.get(path, None)
+        if module is None: # raise if ext is not loaded
+            raise ExtensionError(f"This extension is not loaded ({path})")
+        for name, cog in self.__cogs.copy().items():
+            if _is_submodule(module.__name__, cog.__module__):
+                self.remove_console_cog(name)
+
+        sys.modules.pop(module.__name__, None)  # Remove "cached" module
+
+    def reload_extension(self, path):
+        module = self.__extensions.get(path, None)
+        sys.modules.pop(module.__name__, None)
+        if module is None:
+            raise ExtensionError(f"This extension is not loaded ({path})")
+        old_modules = {}
+        cached = []
+        """
+        Store old module state to fallback if exception occurs
+        """
+        for name, mod in sys.modules.items():
+            if _is_submodule(mod.__name__, path):
+                cached.append(mod)
+        old_modules.update({path: cached})
+
+        try:
+            self.unload_extension(path)
+            self.load_extension(path)
+        except:
+            #  Rollback
+            module.setup(self)
+            self.__extensions[path] = module
+            sys.modules.update(old_modules)
+            raise
 
     def listen(self):
         """
@@ -50,6 +103,8 @@ class Console:
         while True:
             try:
                 console_in = shlex.split(self.input.readline())
+                if len(console_in) == 0:
+                    return
                 try:
                     command = self.__commands__.get(console_in[0], None)
 
@@ -103,15 +158,20 @@ class Console:
             }
         )
 
-    def start(self, loop=None):
+    def remove_command(self, command):
+        self.__commands__.pop(command.name, None)
+
+    def start(self):
         """
-        Abstracts the executor initialization away from you.
-        Takes an asyncio Eventloop as param.
-        :param loop:
+        Abstracts Thread initialization away from user.
         :return:
         """
-        loop = self.client.loop if not loop else loop
-        loop.run_in_executor(None, self.listen)
+        thread = threading.Thread(None, self.listen, daemon=True)
+        thread.start()
+
+
+def _is_submodule(parent, child):
+    return parent == child or child.startswith(parent + ".")
 
 
 class Command:
@@ -204,17 +264,33 @@ def command(**kwargs):
 
 class Cog:
     def __new__(cls, *args, **kwargs):
-        commands = {}
+        commands = []
         # noinspection PyUnresolvedReferences
         for base in reversed(cls.__mro__):
             for elem, value in base.__dict__.items():
                 if isinstance(value, Command):
-                    commands.update({value.name: value})
+                    commands.append(value)
         cls.commands = commands
         return super().__new__(cls)
 
-    def __init__(self):
-        for command in self.commands.values():  # updates the cog attribute for  all commands
-            for c in command.__subcommands__.values():
+    def load(self, console: Console):
+        """
+        Gets called everytime when the Cog gets loaded from console
+        :param console:
+        :return:
+        """
+        for cmd in self.__class__.commands:
+            cmd.cog = self
+            for c in cmd.__subcommands__.values():
                 c.cog = self
-            command.cog = self
+            console.add_command(cmd)
+
+    def unload(self, console: Console):
+        """
+        Gets called when unloaded from console.
+        Cleans up all commands
+        :param console:
+        :return:
+        """
+        for cmd in self.__class__.commands:
+            console.remove_command(cmd)
